/* ReSCAL - Lattice gas
 *
 * Copyright (C) 2011
 *
 * Author: Olivier Rozier <rozier@ipgp.fr>
 *
 * This file is part of ReSCAL.
 *
 * ReSCAL is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include <math.h>
#include <omp.h>
#ifdef __AVX__
#include <stdio.h>
#include <immintrin.h>
#endif

#include "defs.h"
#include "macros.h"
#include "param.h"
#include "cells.h"
#include "space.h"
#include "surface.h"
#include "transitions.h"
#include "lgca.h"
#include "trace.h"
#include "simul.h"

//#undef CYCLAGE_HOR	//pas de cyclage sur le flux (optionnel) -> sinon essoufflement du flux (frottements)

#ifdef LGCA

//#ifdef MVT_VFOR
#if defined(CYCLAGE_HOR) || defined(MVT_VFOR) || defined(MVT_GFOR)
//#define CYCLAGE_HOR //mode MVT_VFOR => cyclage horizontal sur le flux
#define CYCLAGE_MVT
#endif

extern int pbc_mode;  //periodic boundary conditions
extern int boundary;  //boundary conditions
extern char *rot_map;        // periodic mapping of the rotating space
extern Pos2 *rot_map_pos;    // periodic mapping of the rotating space

#ifdef MVT_VFOR
float intensite_vfor = 0.5; //instensite du forcage vertical
#endif

#ifdef MVT_GFOR
float intensite_gfor = 0.001; //instensite du forcage global
#endif


extern unsigned char opt_log;
extern int H, L, D, HL, HLD;       // les dimensions de la terre
extern int LN, LS, LNS, HLN;    //couloir est-ouest (limite nord, limite sud, largeur nord-sud ...)
extern Cell  *TE;	           // la 'terre'
extern int Ncel[];                       // nombre de cellules par type
extern unsigned char opt_vel;       //affichage des vitesses moyennes
extern Vec2 *norm2d;              //normale a la surface
//extern float *norm2dx;          // normale a la surface  (composante X)
//extern float *norm2dy;          // normale a la surface  (composante Y)
extern int h_ceil;  //epaisseur du plafond
extern char *psol;                // indicateur des plans solides verticaux est-ouest
extern const unsigned char Phase[MAX_CELL];	//phase (fluide ou solide) des types de cellules

LgcaSpace Lsv; /// vertical LGCA space
LgcaSpace Lsh; /// horizontal LGCA space
MvtField Collisions[SIZE_MVT_FIELD];  //tableau des collisions fuilde-fluide
#ifdef SOLID_SLIP
MvtField Collisions_SolSlow[16][SIZE_MVT_FIELD]; //tableau des collisions fluide-solide (cas d'une unique particule lente)
MvtField Collisions_SolFast[16][SIZE_MVT_FIELD]; //tableau des collisions fluide-solide (cas d'une unique particule rapide)
#endif
int CelVelx[SIZE_MVT_FIELD];  //precomputed sums of horizontal velocities for a single node
int CelVely[SIZE_MVT_FIELD];  //precomputed sums of vertical velocities for a single node

VelSpace Vsv; /// vertical velocity space
VelSpace Vsh; /// horizontal velocity space

unsigned char NbMvtCell[SIZE_MVT_FIELD];  //nombre de particules en mouvement dans chaque etat d'une cellule
int lgca_iter=0;   //nombre de cycles de collisions
char *nom_fic_mvt = NULL; //nom du fichier VELOC (initial)
int lgca_reset = 1;
int lgca_speedup = 0; //initial speedup of the stabilization of lattice gas
int lgca_mixing = 1; //mixing of the flux (on the left boundary)
int lgca_hplanes = 0; //lattice gas is performed simultaneously on vertical and horizontal planes
char *lgca_boundary_str = NULL;
int lgca_boundary = -1;
int lgca_pbc_mode = -1;

void lecture_mvt();
void do_force_mvt();
void do_mixing();
void compute_vel_interp();
//void do_dump_mvt(int inter);
void dump_densite();
void dump_mvt_in_out();
void dump_collisions();


void params_lgca()
{
  param_family("LGCA", "Lattice-gas parameters");

  /// MVT file
  parameter("Hpp_file", "initial lattice gas - optional", &nom_fic_mvt, PARAM_STRING, "LGCA");
  parameter_hidden("Mvt_file", "Mvt_file parameter is obsolete - Please use Hpp_file parameter instead !", &nom_fic_mvt, PARAM_STRING, "LGCA");

#ifdef MVT_GFOR
  parameter("Lgca_gfor", "global forcing of the flow - optional", &intensite_gfor, PARAM_FLOAT, "LGCA");
  parameter_hidden("Mvt_gfor", "Mvt_gfor parameter is obsolete - Please use Lgca_gfor parameter instead !", &intensite_gfor, PARAM_FLOAT, "LGCA");
#endif
  parameter("Lgca_reset", "reinitialize the flow after each rotation or translation (YES|NO) - optional", &lgca_reset, PARAM_BOOLEAN, "LGCA");
  parameter("Lgca_speedup", "initial speedup of the stabilization of lattice gas - optional", &lgca_speedup, PARAM_INT, "LGCA");
  parameter_hidden("Lgca_hplanes", "lattice gas is performed simultaneously on vertical and horizontal planes (YES|NO) - NO by default", &lgca_hplanes, PARAM_BOOLEAN, "LGCA");
  parameter("Lgca_boundary", "boundary conditions on lattice gas with FULL rotations (FLAT|PERIODIDC) - optional", &lgca_boundary_str, PARAM_STRING, "LGCA");
}

void lgca_parse()
{
  if (lgca_boundary_str){
    if (!strcmp(lgca_boundary_str, "PERIODIC")){
      lgca_boundary = LGCA_BC_PERIODIC;
    }
    else if (!strcmp(lgca_boundary_str, "FLAT")){
      lgca_boundary = LGCA_BC_FLAT;
    }
    else{
      ErrPrintf("Incorrect value for Lgca_boundary : %s\n", lgca_boundary_str);
      exit(-2);
    }
    lgca_pbc_mode = (lgca_boundary == LGCA_BC_PERIODIC);
  }
}

void init_lgca_space(LgcaSpace *p_ls, int orientation, int lx, int ly, int npmax)
{
  /// set dimensions of LGCA planes
  p_ls->orientation = orientation;
  p_ls->lx = lx;
  p_ls->ly = ly;
  p_ls->npmax = npmax;
  p_ls->lxy = lx*ly;
  p_ls->sz = npmax*p_ls->lxy;
  p_ls->pmin = 0;
  p_ls->pmax = npmax-1;
  p_ls->np = npmax;
  p_ls->density = 0;

  LogPrintf("dimensions of %s LGCA space:\n", (orientation==VERTICAL)? "vertical" : "horizontal" );
  LogPrintf("lx = %d\n", p_ls->lx);
  LogPrintf("ly = %d\n", p_ls->ly);
  LogPrintf("npmax = %d\n", p_ls->npmax);
  LogPrintf("pmin = %d\n", p_ls->pmin);
  LogPrintf("pmax = %d\n", p_ls->pmax);

  /// allocation of LGCA buffers
  int ch = (orientation==VERTICAL)? 'v' : 'h';
  char str[20];

  sprintf(str, "Ls%c.mvt", ch);
  AllocMemoryPrint(str, p_ls->mvt, MvtField, p_ls->sz);
  ResetMemory(p_ls->mvt, MvtField, p_ls->sz);

  sprintf(str, "Ls%c.mvt2", ch);
  AllocMemoryPrint(str, p_ls->mvt2, MvtField, p_ls->sz);
  ResetMemory(p_ls->mvt2, MvtField, p_ls->sz);
}

void init_velocity_space(VelSpace *p_vs, LgcaSpace *p_ls)
{
  /// set dimensions of velocity plane
  p_vs->orientation = p_ls->orientation;
  p_vs->lx = (p_ls->lx + VSTEP_X - 1)/VSTEP_X;
  p_vs->ly = (p_ls->ly + VSTEP_Y - 1)/VSTEP_Y;
  p_vs->lxy = p_vs->lx*p_vs->ly;
  p_vs->npmax = p_ls->npmax;
  p_vs->sz = p_vs->lxy*p_vs->npmax;
  p_vs->maxvel = 0;
  p_vs->meanvel = 0;

  LogPrintf("dimensions of %s velocity space:\n", (p_vs->orientation==VERTICAL)? "vertical" : "horizontal" );
  LogPrintf("lx = %d\n", p_vs->lx);
  LogPrintf("ly = %d\n", p_vs->ly);
  LogPrintf("npmax = %d\n", p_vs->npmax);

  /// allocation of velocity planes
  int ch = (p_vs->orientation==VERTICAL)? 'v' : 'h';
  char str[20];

  sprintf(str, "Vs%c.velx", ch);
  AllocMemoryPrint(str, p_vs->velx, int, p_vs->sz);
  ResetMemory(p_vs->velx, int, p_vs->sz);

  sprintf(str, "Vs%c.vely", ch);
  AllocMemoryPrint(str, p_vs->vely, int, p_vs->sz);
  ResetMemory(p_vs->vely, int, p_vs->sz);

  if (VSTEP_TIME > 1){
    sprintf(str, "Vs%c.velx_time", ch);
    AllocMemoryPrint(str, p_vs->velx_time, int, p_vs->sz*VSTEP_TIME);
    ResetMemory(p_vs->velx_time, int, p_vs->sz*VSTEP_TIME);
  }
  else{
    p_vs->velx_time = p_vs->velx;
  }

  if (VSTEP_TIME > 1){
    sprintf(str, "Vs%c.vely_time", ch);
    AllocMemoryPrint(str, p_vs->vely_time, int, p_vs->sz*VSTEP_TIME);
    ResetMemory(p_vs->vely_time, int, p_vs->sz*VSTEP_TIME);
    PrintTotalMemory();
  }
  else{
    p_vs->vely_time = p_vs->vely;
  }
}

void init_lgca()
{
  int i;

  lgca_parse();

  LogPrintf("init_lgca\n");

  LogPrintf("NB_MVT_EO = %d\n", NB_MVT_EO);
  LogPrintf("DIST_MVT_NS = %d\n", DIST_MVT_NS);
  LogPrintf("VSTEP_X = %d\n", VSTEP_X);
  LogPrintf("VSTEP_Y = %d\n", VSTEP_Y);
  LogPrintf("VSTEP_TIME = %d\n", VSTEP_TIME);

  init_lgca_space(&Lsv, VERTICAL, L*NB_MVT_EO, H, (LNS+DIST_MVT_NS-1)/DIST_MVT_NS);
  init_velocity_space(&Vsv, &Lsv);

  /*if (lgca_hplanes){
    init_lgca_space(&Lsh, HORIZONTAL, L*NB_MVT_EO, D, (H+DIST_MVT_H-1)/DIST_MVT_H);
    init_velocity_space(&Vsh, &Lsh);
  }*/

#if (VSTEP == 1)
  LogPrintf("VMEAN_REF = %d\n", VMEAN_REF);
#endif
#ifdef MVT_HFOR
  LogPrintf("forcage flux par tapis roulant\n");
#endif
#ifdef MVT_VFOR
  LogPrintf("forcage flux sur le bord gauche, intensite = %f\n", intensite_vfor);
#endif
#ifdef MVT_GFOR
  LogPrintf("forcage global du flux, intensite = %f\n", intensite_gfor);
#endif
  lgca_mixing = !pbc_mode; //mixing when cellular space is not periodic
  if (lgca_mixing) LogPrintf("brouillage flux sur le bord gauche\n");
#ifdef SOLID_SLIP
  LogPrintf("slip boundary condition : SOLID_SLIP\n");
#endif
//#ifdef SURFACE_SLIP
//  LogPrintf("slip boundary condition : SURFACE_SLIP\n");
//#endif
#ifdef CEIL_SLIP
  LogPrintf("slip boundary condition : CEIL_SLIP\n");
//  LogPrintf("epaisseur du plafond = %d\n", h_ceil);
#endif
//#ifdef VEL_SLIDE
//  LogPrintf("sliding window for the averaging of the velocities\n");
//#endif

  PrintTotalMemory();

  if (nom_fic_mvt){
    lecture_mvt();
  }
  else{
    init_mvt();
  }

  // initialisation tableau NbMvtCell[]
  for (i=0; i<SIZE_MVT_FIELD; i++){
    NbMvtCell[i] = 0;
    if (i & MVT_E) NbMvtCell[i]++;
    if (i & MVT_O) NbMvtCell[i]++;
    if (i & MVT_B) NbMvtCell[i]++;
    if (i & MVT_H) NbMvtCell[i]++;
    if (i & MVT_EB) NbMvtCell[i]++;
    if (i & MVT_EH) NbMvtCell[i]++;
    if (i & MVT_OB) NbMvtCell[i]++;
    if (i & MVT_OH) NbMvtCell[i]++;
  }

  // initialisation tableau Collisions[]
  for (i=0; i<SIZE_MVT_FIELD; i++){
    if (i & MVT_OUT){
      Collisions[i] = MVT_OUT;
    }
    else if (i & MVT_SOLID){
      Collisions[i] = MVT_SOLID;
      if (i & MVT_E) Collisions[i] |= MVT_O;
      //if (i & MVT_E) Collisions[i] |= MVT_H; //test
      if (i & MVT_O) Collisions[i] |= MVT_E;
      if (i & MVT_B) Collisions[i] |= MVT_H;
      if (i & MVT_H) Collisions[i] |= MVT_B;

      if (i & MVT_EB) Collisions[i] |= MVT_OH; //MVT_OH;
      if (i & MVT_EH) Collisions[i] |= MVT_OB; //MVT_OB;
      if (i & MVT_OB) Collisions[i] |= MVT_EH; //MVT_EH;
      if (i & MVT_OH) Collisions[i] |= MVT_EB; //MVT_EB;

      /*if (i & MVT_EB) Collisions[i] |= MVT_EH; //MVT_OH;
      if (i & MVT_EH) Collisions[i] |= MVT_EB; //MVT_OB;
      if (i & MVT_OB) Collisions[i] |= MVT_OH; //MVT_EH;
      if (i & MVT_OH) Collisions[i] |= MVT_OB; //MVT_EB;*/

      /*if (i & MVT_EB) Collisions[i] |= MVT_OB; //MVT_OH;
      if (i & MVT_EH) Collisions[i] |= MVT_EB; //MVT_OB;
      if (i & MVT_OB) Collisions[i] |= MVT_OH; //MVT_EH;
      if (i & MVT_OH) Collisions[i] |= MVT_EH; //MVT_EB;*/
    }
    else{
      Collisions[i] = 0;

      if ((i & MVT_SLOW) == MVT_E){
        if ((i & MVT_FAST) == MVT_OB) Collisions[i] |= (MVT_O | MVT_EB);
	else if ((i & MVT_FAST) == MVT_OH) Collisions[i] |= (MVT_O | MVT_EH);
	else Collisions[i] |= MVT_E;
      }
      else if ((i & MVT_SLOW) == MVT_O){
        if ((i & MVT_FAST) == MVT_EB) Collisions[i] |= (MVT_E | MVT_OB);
	else if ((i & MVT_FAST) == MVT_EH) Collisions[i] |= (MVT_E | MVT_OH);
	else Collisions[i] |= MVT_O;
      }
      else if ((i & MVT_SLOW) == MVT_B){
        if ((i & MVT_FAST) == MVT_EH) Collisions[i] |= (MVT_H | MVT_EB);
	else if ((i & MVT_FAST) == MVT_OH) Collisions[i] |= (MVT_H | MVT_OB);
	else Collisions[i] |= MVT_B;
      }
      else if ((i & MVT_SLOW) == MVT_H){
        if ((i & MVT_FAST) == MVT_EB) Collisions[i] |= (MVT_B | MVT_EH);
	else if ((i & MVT_FAST) == MVT_OB) Collisions[i] |= (MVT_B | MVT_OH);
	else Collisions[i] |= MVT_H;
      }
      else if ((i & MVT_SLOW) == (MVT_E | MVT_O)) Collisions[i] |= (MVT_H | MVT_B);
      else if ((i & MVT_SLOW) == (MVT_H | MVT_B)) Collisions[i] |= (MVT_E | MVT_O);
      else Collisions[i] |= (i & MVT_SLOW);

      if ((i & MVT_FAST) == (MVT_EB | MVT_OH)) Collisions[i] |= (MVT_EH | MVT_OB);
      else if ((i & MVT_FAST) == (MVT_EH | MVT_OB)) Collisions[i] |= (MVT_EB | MVT_OH);
      else if (!(Collisions[i] & MVT_FAST)) Collisions[i] |= (i & MVT_FAST);
    }
  }

#ifdef SOLID_SLIP
  // initialisation (d'une partie) du tableau Collisions_SolSlow[]
  int mvt_sol_E, mvt_sol_O, mvt_sol_H, mvt_sol_B, mvt_sol;
  LogPrintf("collisions realistes : solslow\n");
  memset(Collisions_SolSlow,0,sizeof(Collisions_SolSlow));
  for (mvt_sol=0; mvt_sol<16; mvt_sol++){
    mvt_sol_E = mvt_sol & (MVT_E >> 2);
    mvt_sol_O = mvt_sol & (MVT_O >> 2);
    mvt_sol_B = mvt_sol & (MVT_B >> 2);
    mvt_sol_H = mvt_sol & (MVT_H >> 2);

    Collisions_SolSlow[mvt_sol][MVT_E] = (!mvt_sol_H) ? MVT_H : (!mvt_sol_B) ? MVT_B : MVT_O;
    Collisions_SolSlow[mvt_sol][MVT_O] = (!mvt_sol_H) ? MVT_H : (!mvt_sol_B) ? MVT_B : MVT_E;
    Collisions_SolSlow[mvt_sol][MVT_B] = (!mvt_sol_E) ? MVT_E : (!mvt_sol_O) ? MVT_O : MVT_H;
    //Collisions_SolSlow[mvt_sol][MVT_B] = MVT_H;
    Collisions_SolSlow[mvt_sol][MVT_H] = (!mvt_sol_E) ? MVT_E : (!mvt_sol_O) ? MVT_O : MVT_B;
  }

  // initialisation (d'une partie) du tableau Collisions_SolFast[]
  int mvt_sol_EB, mvt_sol_OB, mvt_sol_EH, mvt_sol_OH;//, mvt_sol;
  LogPrintf("collisions realistes : solfast\n");
  memset(Collisions_SolFast,0,sizeof(Collisions_SolFast));
  for (mvt_sol=0; mvt_sol<16; mvt_sol++){
    mvt_sol_EB = mvt_sol & (MVT_EB >> 6);
    mvt_sol_EH = mvt_sol & (MVT_EH >> 6);
    mvt_sol_OB = mvt_sol & (MVT_OB >> 6);
    mvt_sol_OH = mvt_sol & (MVT_OH >> 6);

    Collisions_SolFast[mvt_sol][MVT_EB] = (!mvt_sol_EH) ? MVT_EH : (!mvt_sol_OB) ? MVT_OB : MVT_OH;
    Collisions_SolFast[mvt_sol][MVT_EH] = (!mvt_sol_OH) ? MVT_OH : (!mvt_sol_EB) ? MVT_EB : MVT_OB;
    Collisions_SolFast[mvt_sol][MVT_OB] = (!mvt_sol_OH) ? MVT_OH : (!mvt_sol_EB) ? MVT_EB : MVT_EH;
    Collisions_SolFast[mvt_sol][MVT_OH] = (!mvt_sol_EH) ? MVT_EH : (!mvt_sol_OB) ? MVT_OB : MVT_EB;
    //Collisions_SolFast[mvt_sol][MVT_EB] = MVT_EH;
    //Collisions_SolFast[mvt_sol][MVT_EH] = MVT_EB;
    //Collisions_SolFast[mvt_sol][MVT_OB] = MVT_OH;
    //Collisions_SolFast[mvt_sol][MVT_OH] = MVT_OB;
  }
#endif

  if (opt_log) dump_collisions();

  // initialization of precomputed sums of velocities
  int *pvx = CelVelx;
  int *pvy = CelVely;
  for(i=0; i<SIZE_MVT_FIELD; i++, pvx++, pvy++){
    if (i & MVT_SLOW){
      if (i & MVT_E) (*pvx)++;
      if (i & MVT_O) (*pvx)--;
      if (i & MVT_B) (*pvy)++;
      if (i & MVT_H) (*pvy)--;
    }
    if (i & MVT_FAST){
      if (i & MVT_EB) {(*pvx)++; (*pvy)++;}
      if (i & MVT_EH) {(*pvx)++; (*pvy)--;}
      if (i & MVT_OB) {(*pvx)--; (*pvy)++;}
      if (i & MVT_OH) {(*pvx)--; (*pvy)--;}
    }
  }

  //dump_densite();

#ifdef DUMP_SIGNATURE
  //empreinte gazeuse
  if (opt_log) dump_signature_mvt();
#endif

#ifdef OPENMP
#pragma omp parallel
  {
    int tid = omp_get_thread_num();
    if (tid==0){
      int nthreads = omp_get_num_threads();
      LogPrintf("number of OMP threads : %d\n", nthreads);
    }
  }
#endif
}


// initialize one site of lattice gas
void init_mvt_cell(MvtField *mvt)
{
  static MvtField mvt_slow[] = {MVT_E, MVT_O, MVT_B, MVT_H};
  static MvtField mvt_fast[] = {MVT_EB, MVT_EH, MVT_OB, MVT_OH};

  int alea;
  //if (drand48() < 0.5)
  /*if (j>0.6*H)*/{
    /* *mvt = MVT_E;
    nb_mvt++;
    float aleat = drand48();
    if (aleat < 0.25) {*mvt |= MVT_EB; nb_mvt++;}
    else if (aleat < 0.5) {*mvt |= MVT_EH; nb_mvt++;}*/

    // 1 particule
    /*int alea8=drand48()*8;
    *aux_mvt = (alea8<1) ? MVT_E : 0;
    // *aux_mvt = (alea8<4) ? mvt_slow[alea8] : (alea8<8) ? mvt_fast[alea8-4] : 0;
            if (*aux_mvt) nb_mvt_in++;*/

    // 1 particule lente
    alea=drand48()*4;
    if (alea<4){
      *mvt = mvt_slow[alea];
    }
    //*aux_mvt = MVT_E;
    //*aux_mvt = (alea4) ? MVT_E : 0;

    // 1 particule rapide
    //alea4=drand48()*4;
    //*aux_mvt |= mvt_fast[alea4];

    // 1/2 particule rapide
    alea=drand48()*8;
    if (alea<4){ // 1/2 particule rapide
      *mvt |= mvt_fast[alea];
    }

    // 2 particules lentes
    /*int alea2=drand48()*2;
    *aux_mvt = mvt_slow[alea2]; // 1 particule lente horizontale
    alea2=drand48()*2;
    *aux_mvt |= mvt_slow[2+alea2]; // 1 particule lente verticale
    nb_mvt_in +=2;*/

    // 2 particules rapides
    /*alea2=drand48()*2;
    *aux_mvt |= mvt_fast[alea2]; // 1 particule rapide vers le bas
    alea2=drand48()*2;
    *aux_mvt |= mvt_fast[2+alea2]; // 1 particule rapide vers le haut
    nb_mvt_in +=2;*/

    // 6 particules
    /*int alea4=drand48()*4;
    *aux_mvt = MVT_SLOW ^ mvt_slow[alea4]; // 3 particules lentes
    alea4=drand48()*4;
    *aux_mvt |= MVT_FAST ^ mvt_fast[alea4]; // 3 particules rapides
    nb_mvt_in +=6;*/
  }
}

/// initialization of lgca planes
void init_mvt()
{
  int i, j, k, ic;
  Cell *aux;
  MvtField *aux_mvt;

  LogPrintf("reset lattice gas\n");

  if (rot_map){
    /// set the north-south boundaries (we have to keep an extra vertical plan on north and south sides for the interpolation)
    Lsv.pmin = 0;
    for(k=0; psol[LN + k*DIST_MVT_NS]; k++) Lsv.pmin = k;
    for(; !psol[LN + k*DIST_MVT_NS]; k++);
    Lsv.pmax = k;
    Lsv.np = Lsv.pmax-Lsv.pmin+1;
    assert(Lsv.pmax < Lsv.npmax);
  }
  //LogPrintf("init_mvt: Lsv.pmin=%d, Lsv.pmax=%d\n", Lsv.pmin, Lsv.pmax);

  // initialisation tableau Lsv.mvt[]
  Lsv.nfc = 0;
#if defined(OPENMP) && !defined(DETERMINISTIC)
#pragma omp parallel for private(i,j,k,ic,aux,aux_mvt) reduction(+:Lsv.nfc)
#endif
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ //profondeur
    aux = TE + HLN + k*HL*DIST_MVT_NS;
    aux_mvt = Lsv.mvt + k*Lsv.lxy;
    for(j=0; j<H; j++){ //hauteur
      for(i=0; i<L; i++, aux++){ //largeur
        for(ic=0; ic<NB_MVT_EO; ic++, aux_mvt++){
          if (Phase[aux->celltype] == FLUID){
            *aux_mvt=0;
            Lsv.nfc++;
            init_mvt_cell(aux_mvt);
          }
          else if (Phase[aux->celltype] == SOLID)
            *aux_mvt = MVT_SOLID;

          //if ((j==0) || (j==H-1)) //bords horizontaux
          //if ((j==1) || (j==H-2)) //bords horizontaux
          //  *aux_mvt = MVT_SOLID;
#if !defined(CYCLAGE_MVT)
          else if (i==L-1) // bord est
            *aux_mvt = MVT_OUT;
#endif
        }
      }
    }
  }

  if (rot_map) out_of_space_mvt(1);

  //speed up the stabilization of the flow
  if (lgca_speedup){
    //int nb_cyc_for=VSTEP_TIME*10;
    LogPrintf("speedup of the lattice gas: %d\n", lgca_speedup);
    for(i=0; i<lgca_speedup; i++) do_force_mvt();
  }

  if (Vsv.velx_interp){
    ResetMemory(Vsv.velx_interp, float, HLD);
    ResetMemory(Vsv.vely_interp, float, HLD);
  }
  if (Vsv.velx_interp){
    ResetMemory(Vsv.velx_interp, float, HLD);
    ResetMemory(Vsv.velx_interp, float, HLD);
  }
}


/// set the lattice gas out of the rotating space, using periodic boundary conditions
void out_of_space_mvt(int reset_mvt)
{
  int i, j, k, ic, ck, typ;
  MvtField *aux_mvt;
  Pos2 cp;

  if (lgca_pbc_mode < 0) lgca_pbc_mode = pbc_mode;
  if (reset_mvt) LogPrintf("initialization of the lattice gas outside the rotating space with %s topography\n", (lgca_pbc_mode ? "periodic" : "flat"));

#if defined(OPENMP) && !defined(DETERMINISTIC)
#pragma omp parallel for private(i,j,k,ck,ic,aux_mvt,cp,typ) reduction(+:Lsv.nfc)
#endif
  for(ck=Lsv.pmin; ck<=Lsv.pmax; ck++){ //profondeur
    k = LN + ck*DIST_MVT_NS;
    for(j=0; j<H; j++){ //hauteur
      aux_mvt = Lsv.mvt + j*Lsv.lx + ck*Lsv.lxy;
      for(i=0; i<L; i++){ //largeur
        if (OutOfSpace(i, k)){
          if (lgca_pbc_mode){
            /// Periodic boundary conditions: we replicate the topography periodically
            cp = RotMapPos(i, k);
            typ = CellType(cp.x, j, cp.y);
            for(ic=0; ic<NB_MVT_EO; ic++, aux_mvt++){
              if (Phase[typ] == SOLID)
                SetMask(*aux_mvt, MVT_SOLID);
              else{
                UnsetMask(*aux_mvt, MVT_SOLID);
                if (reset_mvt){
                  init_mvt_cell(aux_mvt);
                  Lsv.nfc++;
                }
              }
            }
          }
          /*else if (boundary == BC_CLOSE){
            for(ic=0; ic<NB_MVT_EO; ic++, aux_mvt++){
              SetMask(*aux_mvt, MVT_SOLID);
            }
          }*/
          else{
            /// Flat boundary conditions:  we put solid ground and ceiling
            for(ic=0; ic<NB_MVT_EO; ic++, aux_mvt++){
              if ((j<=1) || (j>=H-2)){
                SetMask(*aux_mvt, MVT_SOLID);
              }
              else{
                UnsetMask(*aux_mvt, MVT_SOLID);
                if (reset_mvt){
                  init_mvt_cell(aux_mvt);
                  Lsv.nfc++;
                }
              }
            }
          }
        }
        else aux_mvt += NB_MVT_EO;
      }
    }
  }
}



// modification globale de la terre
void collisions_mod_terre()
{
  int i, j, k, ic;
  Cell *aux;
  MvtField *aux_mvt;

  Lsv.nfc = 0;
  aux = TE+HLN;
  aux_mvt = Lsv.mvt;
  for(k=Lsv.pmin; k<=Lsv.pmax; k++, aux+=HL*(DIST_MVT_NS-1)){ //profondeur
    for(j=0; j<H; j++){ //hauteur
      for(i=0; i<L; i++, aux++){ //largeur
        for(ic=0; ic<NB_MVT_EO; ic++, aux_mvt++){
          if (Phase[aux->celltype] == SOLID){
            SetMask(*aux_mvt, MVT_SOLID);
          }
          else{
            UnsetMask(*aux_mvt, MVT_SOLID);
            Lsv.nfc++;
          }
        }
      }
    }
  }
}

void lecture_mvt()
{
  FILE *fp;
  int index;

  LogPrintf("lecture fichier HPP : %s\n",nom_fic_mvt);

  fp = fopen(nom_fic_mvt,"r");
  if ( ! fp ){
    ErrPrintf("ERROR: cannot open HPP file %s\n", nom_fic_mvt);
    exit(-4);
  }

  fread(Lsv.mvt, sizeof(MvtField), Lsv.sz, fp);

  fclose(fp);

  /// initialization of nfc field
  for (index=0; index<Lsv.sz; index++){
    if (!(Lsv.mvt[index] & MVT_SOLID)) Lsv.nfc++;
  }

  /// initial velocity space
  /*for (i=0; i<VSTEP_TIME; i++)*/ compute_vel(0);
}


int mvt(int index)
{
  //return (Lsv.mvt[index] & (MVT_E | MVT_O | MVT_B | MVT_H));
  return (Lsv.mvt[index] & MVT_ALLDIR);
  //return (NbMvtCell[Lsv.mvt[index]] > Lsv.density);
}


int mvt_solid(int ii)
{
  return (Lsv.mvt[ii] & MVT_SOLID);
}

int mvt_bas(int ii)
{
  return (Lsv.mvt[ii] & (MVT_B | MVT_EB | MVT_OB));
}

int mvt_haut(int ii)
{
  return (Lsv.mvt[ii] & (MVT_H | MVT_EH | MVT_OH));
}

int mvt_est(int ii)
{
  //return (Lsv.mvt[ii] & (MVT_EB | MVT_EH));
  return (Lsv.mvt[ii] & (MVT_E | MVT_EB | MVT_EH));
}

int mvt_ouest(int ii)
{
  //return (Lsv.mvt[ii] & MVT_O);
  //return (Lsv.mvt[ii] & (MVT_OB | MVT_OH));
  return (Lsv.mvt[ii] & (MVT_O | MVT_OB | MVT_OH));
}

int check_mvt(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & MVT_ALLDIR);
}

int check_mvt_solid(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & MVT_SOLID);
}

int check_mvt_bas(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & (MVT_B | MVT_EB | MVT_OB));
}

int check_mvt_haut(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & (MVT_H | MVT_EH | MVT_OH));
}

int check_mvt_est(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & (MVT_E | MVT_EB | MVT_EH));
  //return (Lsv.mvt[ii] & MVT_E);
}

int check_mvt_ouest(int index, void *data)
{
  int ii;
  Calcule_cix(index, ii);
  return (Lsv.mvt[ii] & (MVT_O | MVT_OB | MVT_OH));
}

int check_mvt_est_et_bas(int index, void *data)
{
  return check_mvt_est(index, NULL) && check_mvt_bas(index, NULL);
}


int check_no_mvt_est_et_bas(int index, void *data)
{
  return !check_mvt_est_et_bas(index, NULL);
}

void lgca_modify_cell(int type, int index)
{
  //int ii=(index-HLN)*NB_MVT_EO;
  int x, y, z, ii, ic, cz;
  Calcule_xyz(index, x, y, z);
  cz = (z-LN)/DIST_MVT_NS;
  if (z != cz*DIST_MVT_NS + LN){
    //ErrPrintf("lgca_modify_cell : mauvais plan ! (%d)\n", z);
    return;
  }
  ii = x*NB_MVT_EO + y*Lsv.lx + cz*Lsv.lxy;
  //Calcul_cix(index, ii);

  for(ic=0; ic<NB_MVT_EO; ic++, ii++){
    if (Phase[type] == SOLID){
      if (!(Lsv.mvt[ii] & MVT_SOLID)) Lsv.nfc--;
      SetMask(Lsv.mvt[ii], MVT_SOLID);
      //if (mvt(index)) {LogPrintf("vert piege !!\n");}
    }
    else{
      if (Lsv.mvt[ii] & MVT_SOLID) Lsv.nfc++;
      UnsetMask(Lsv.mvt[ii], MVT_SOLID);
    }
  }
}


void lgca_collisions()
{
  int i, j, k, ii, ii2;

#ifdef OPENMP
#pragma omp parallel for private(i,ii,j,k)
#endif
//shared(Lsv.np, Lsv.lx, Lsv.ly, Collisions, Lsv.mvt, Lsv.mvt2, NbMvtCell, h_ceil)
  //LogPrintf("lgca_collisions\n");
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ // profondeur
    /* Obtain thread number */
    //int tid = omp_get_thread_num();

    ii = k*Lsv.lxy + Lsv.lx;
    //LogPrintf("tid=%d k=%d ii=%d\n", tid, k, ii);
    //ii+=Lsv.lx;
    for(j=1; j<Lsv.ly-1; j++){ // hauteur
      ii += NB_MVT_EO;
      for(i=NB_MVT_EO; i<Lsv.lx-NB_MVT_EO; i++, ii++){ //largeur
        Lsv.mvt2[ii] = Collisions[Lsv.mvt[ii]];
        //Lsv.mvt2[ii] = Lsv.mvt[ii];

#ifdef CEIL_SLIP
        if ((j <= h_ceil) && (NbMvtCell[Lsv.mvt[ii] & MVT_FAST]==1)){
          //rebond realiste au plafond
          MvtField mvt2 = 0;
          switch (Lsv.mvt[ii] & MVT_FAST)
          {
            case MVT_EH:
              mvt2 = MVT_EB;
              break;
            case MVT_OH:
              mvt2 = MVT_OB;
              break;
          }
          if (mvt2){
            UnsetMask(Lsv.mvt2[ii], MVT_FAST);
            SetMask(Lsv.mvt2[ii], mvt2);
          }
        }
#endif //CEIL_SLIP

#ifdef SOLID_SLIP
        if (mvt_solid(ii) && (NbMvtCell[Lsv.mvt[ii] & MVT_SLOW]==1)){
          //EST
          ii2 = ii+1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==Lsv.lx-NB_MVT_EO-1)) ii2 -= Lsv.lx-2*NB_MVT_EO;
#endif
          //OUEST
          int mvt_sol_E = mvt_solid(ii2) ? (MVT_E >> 2) : 0;
          ii2 = ii-1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==NB_MVT_EO)) ii2 += Lsv.lx-2*NB_MVT_EO;
#endif
          int mvt_sol_O = mvt_solid(ii-1) ? (MVT_O >> 2) : 0;
          //BAS
          int mvt_sol_B = mvt_solid(ii+Lsv.lx) ? (MVT_B >> 2) : 0;
          //HAUT
          int mvt_sol_H = mvt_solid(ii-Lsv.lx) ? (MVT_H >> 2) : 0;
          int mvt_sol = mvt_sol_E | mvt_sol_O | mvt_sol_B | mvt_sol_H;
          UnsetMask(Lsv.mvt2[ii], MVT_SLOW);
          SetMask(Lsv.mvt2[ii], Collisions_SolSlow[mvt_sol][Lsv.mvt[ii] & MVT_SLOW]);
        }

        if (mvt_solid(ii) && (NbMvtCell[Lsv.mvt[ii] & MVT_FAST]==1)){

          //EST-BAS
          ii2 = ii+Lsv.lx+1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==Lsv.lx-NB_MVT_EO-1)) ii2 -= Lsv.lx-2*NB_MVT_EO;
#endif
          int mvt_sol_EB = mvt_solid(ii2) ? (MVT_EB >> 6) : 0;

          //EST-HAUT
          ii2 = ii-Lsv.lx+1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==Lsv.lx-NB_MVT_EO-1)) ii2 -= Lsv.lx-2*NB_MVT_EO;
#endif
          int mvt_sol_EH = mvt_solid(ii2) ? (MVT_EH >> 6) : 0;

          //OUEST-BAS
          ii2 = ii+Lsv.lx-1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==NB_MVT_EO)) ii2 += Lsv.lx-2*NB_MVT_EO;
#endif
          int mvt_sol_OB = mvt_solid(ii2) ? (MVT_OB >> 6) : 0;

          //OUEST-HAUT
          ii2 = ii-Lsv.lx-1;
#ifdef CYCLAGE_HOR
          if (pbc_mode && (i==NB_MVT_EO)) ii2 += Lsv.lx-2*NB_MVT_EO;
#endif
          int mvt_sol_OH = mvt_solid(ii2) ? (MVT_OH >> 6) : 0;
          int mvt_sol = mvt_sol_EB | mvt_sol_EH | mvt_sol_OB | mvt_sol_OH;
          UnsetMask(Lsv.mvt2[ii], MVT_FAST);
          SetMask(Lsv.mvt2[ii], Collisions_SolFast[mvt_sol][Lsv.mvt[ii] & MVT_FAST]);
        }
#endif //SOLID_SLIP
      }
      ii += NB_MVT_EO;
    }
    ii+=Lsv.lx;
  }
  lgca_iter++;
}


void propagate_mvtcel(int i, int ii)
{
  int ii0;

  Lsv.mvt[ii]=0;

  //SOLID+OUT
  Lsv.mvt[ii] |= (Lsv.mvt2[ii] & (MVT_SOLID | MVT_OUT));

  //EST
  ii0 = ii-1;
#ifdef CYCLAGE_MVT
  if (i==NB_MVT_EO) ii0 += Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_E);

  //OUEST
  ii0 = ii+1;
#ifdef CYCLAGE_MVT
  if (i==Lsv.lx-NB_MVT_EO-1) ii0 -= Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_O);

  //BAS
  ii0 = ii-Lsv.lx;
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_B);

  //HAUT
  ii0 = ii+Lsv.lx;
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_H);

  //EST+BAS
  ii0 = ii-Lsv.lx-1;
#ifdef CYCLAGE_MVT
  if (i==NB_MVT_EO) ii0 += Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_EB);

  //EST+HAUT
  ii0 = ii+Lsv.lx-1;
#ifdef CYCLAGE_MVT
  if (i==NB_MVT_EO) ii0 += Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_EH);

  //OUEST+BAS
  ii0 = ii-Lsv.lx+1;
#ifdef CYCLAGE_MVT
  if (i==Lsv.lx-NB_MVT_EO-1) ii0 -= Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_OB);

  //OUEST+HAUT
  ii0 = ii+Lsv.lx+1;
#ifdef CYCLAGE_MVT
  if (i==Lsv.lx-NB_MVT_EO-1) ii0 -= Lsv.lx-2*NB_MVT_EO;
#endif
  Lsv.mvt[ii] |= (Lsv.mvt2[ii0] & MVT_OH);

}

#ifdef USE_AVX
void propagate_mvtcel_vec(int i, int ii)
{
  static int step = sizeof(__m256d)/sizeof(MvtField);
  static int start = 1;
  static __m256i mm_maskv_so, mm_maskv_e, mm_maskv_w, mm_maskv_u, mm_maskv_d, mm_maskv_ue, mm_maskv_uw, mm_maskv_de, mm_maskv_dw;
  __m256i *mm_valp, mm_a, mm_b, mm_res, *mm_curp;

  if (start){
    mm_maskv_so = _mm256_set1_epi16(MVT_SOLID | MVT_OUT);
    mm_maskv_e = _mm256_set1_epi16(MVT_E);
    mm_maskv_w = _mm256_set1_epi16(MVT_O);
    mm_maskv_u = _mm256_set1_epi16(MVT_H);
    mm_maskv_d = _mm256_set1_epi16(MVT_B);
    mm_maskv_ue = _mm256_set1_epi16(MVT_EH);
    mm_maskv_uw = _mm256_set1_epi16(MVT_OH);
    mm_maskv_de = _mm256_set1_epi16(MVT_EB);
    mm_maskv_dw = _mm256_set1_epi16(MVT_OB);
#ifdef OPENMP
    int tid=omp_get_thread_num();
    //LogPrintf("thread_num=%d\n", tid);
    if (tid==0)
#endif
    {
      start = 0;
      LogPrintf("AVX optimization: %d MVT cells per vector\n", step);
    }
  }

  if (!CheckAlign32(Lsv.mvt2+ii)){
    ErrPrintf("ERROR: address of (Lsv.mvt2+ii) is misaligned for AVX (256b)\n");
    exit(-1);
  };


  /// SOLID+OUT
  mm_valp = (__m256i*) (Lsv.mvt2+ii);
  mm_a = _mm256_load_si256(mm_valp);
  mm_res = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_so);

  /// EAST
  mm_valp = (__m256i*) (Lsv.mvt2+ii-1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_e);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// WEST
  mm_valp = (__m256i*) (Lsv.mvt2+ii+1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_w);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// UP
  mm_valp = (__m256i*) (Lsv.mvt2+ii+Lsv.lx);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_u);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// DOWN
  mm_valp = (__m256i*) (Lsv.mvt2+ii-Lsv.lx);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_d);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// UP-EAST
  mm_valp = (__m256i*) (Lsv.mvt2+ii+Lsv.lx-1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_ue);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// UP-WEST
  mm_valp = (__m256i*) (Lsv.mvt2+ii+Lsv.lx+1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_uw);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// DOWN-EAST
  mm_valp = (__m256i*) (Lsv.mvt2+ii-Lsv.lx-1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_de);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// DOWN-WEST
  mm_valp = (__m256i*) (Lsv.mvt2+ii-Lsv.lx+1);
  mm_a = _mm256_loadu_si256(mm_valp);
  mm_b = (__m256i) _mm256_and_pd((__m256d)mm_a, (__m256d)mm_maskv_dw);
  mm_res = (__m256i) _mm256_or_pd((__m256d)mm_b, (__m256d)mm_res);

  /// UPDATE CURRENT VECTOR
  mm_curp = (__m256i*) (Lsv.mvt+ii);
  _mm256_store_si256(mm_curp, mm_res);
}
#endif // USE_AVX

void lgca_propagations()
{
  int i, j, k, ii;
#ifdef USE_AVX
  static int step = sizeof(__m256d)/sizeof(MvtField);
#endif

#ifdef OPENMP
#pragma omp parallel for private(i,j,k,ii)
#endif
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ // profondeur
    for(j=1; j<Lsv.ly-1; j++){ // hauteur
      i = NB_MVT_EO;
      ii = Lsv.lx*Lsv.ly*k + Lsv.lx*j + i;
      while(i<Lsv.lx-NB_MVT_EO){ //largeur
#ifdef USE_AVX
        if (CheckAlign32(Lsv.mvt+ii) && (i>1) && (i+step < Lsv.lx-NB_MVT_EO)){
          propagate_mvtcel_vec(i ,ii);
          i += step;
          ii += step;
        }
        else{
          propagate_mvtcel(i++, ii++);
        }
#else
        propagate_mvtcel(i++, ii++);
#endif // USE_AVX
      }
    }
  }

#if defined(MVT_HFOR) || defined(MVT_VFOR) || defined(MVT_GFOR) //forcage flux
  do_force_mvt();
#endif

  if (lgca_mixing) do_mixing();

#ifdef DUMP_SIGNATURE
  //empreinte gazeuse
  if (opt_log && !(lgca_iter & 0xf)) dump_signature_mvt(); //1 cycle sur 16
#endif
}


void do_force_mvt()
{
  int x, y, z;
  int ix;

#ifdef MVT_HFOR
  // forcage flux ouest-est sur la couche superieure
  //for(y=1; y<Lsv.ly/2; y++)

  for(z=Lsv.pmin; z<=Lsv.pmax; z++){
    for(x=0; x<Lsv.lx; x++){
      ix = x+Lsv.lx*(h_ceil+1)+z*Lsv.lxy;
      if ((Lsv.mvt[ix] & (MVT_E | MVT_O)) == MVT_O){
        UnsetMask(Lsv.mvt[ix], MVT_O);
        SetMask(Lsv.mvt[ix], MVT_E);
      }
      if ((Lsv.mvt[ix] & (MVT_EH | MVT_OH)) == MVT_OH){
        UnsetMask(Lsv.mvt[ix], MVT_OH);
        SetMask(Lsv.mvt[ix], MVT_EH);
      }
      if ((Lsv.mvt[ix] & (MVT_EB | MVT_OB)) == MVT_OB){
        UnsetMask(Lsv.mvt[ix], MVT_OB);
        SetMask(Lsv.mvt[ix], MVT_EB);
      }
    }
  }


  // forcage flux est-ouest sur la couche superieure
  /*  for(x=0; x<L; x++){
      ix = x+L;
      if ((Lsv.mvt[ix] & (MVT_E | MVT_O)) == MVT_E){
        Lsv.mvt[ix] &= (-1) ^ MVT_E;
        Lsv.mvt[ix] |= MVT_O;
      }
      if ((Lsv.mvt[ix] & (MVT_EH | MVT_OH)) == MVT_EH){
        Lsv.mvt[ix] &= (-1) ^ MVT_EH;
        Lsv.mvt[ix] |= MVT_OH;
      }
      if ((Lsv.mvt[ix] & (MVT_EB | MVT_OB)) == MVT_EB){
        Lsv.mvt[ix] &= (-1) ^ MVT_EB;
        Lsv.mvt[ix] |= MVT_OB;
      }
  }*/

  // forcage flux ouest-est sur la couche inferieure
  /*for(x=0; x<L; x++){
    ix = x+HL-2*L;
    if ((Lsv.mvt[ix] & (MVT_E | MVT_O)) == MVT_O){
      Lsv.mvt[ix] &= (-1) ^ MVT_O;
      Lsv.mvt[ix] |= MVT_E;
    }
    if ((Lsv.mvt[ix] & (MVT_EH | MVT_OH)) == MVT_OH){
      Lsv.mvt[ix] &= (-1) ^ MVT_OH;
      Lsv.mvt[ix] |= MVT_EH;
    }
    if ((Lsv.mvt[ix] & (MVT_EB | MVT_OB)) == MVT_OB){
      Lsv.mvt[ix] &= (-1) ^ MVT_OB;
      Lsv.mvt[ix] |= MVT_EB;
    }
  }*/
#endif

#ifdef MVT_VFOR
  // forcage partiel du flux sur le bord gauche
  int cpt_vfor=0;

  for(z=Lsv.pmin; z<=Lsv.pmax; z++){
    for(x=0; x<1; x++){
      ix = z*Lsv.lxy+5*Lsv.lx+NB_MVT_EO+x;
      //for (y=5; y<Lsv.ly-5; y++, ix+=Lsv.lx){
      for (y=0; y<Lsv.ly; y++, ix+=Lsv.lx){
        //if (drand48()<1.0){
        //forcage
        if (drand48() < intensite_vfor){
        //if (1){
          if ((Lsv.mvt[ix] & (MVT_E | MVT_O)) == MVT_O){
            UnsetMask(Lsv.mvt[ix], MVT_O);
            SetMask(Lsv.mvt[ix], MVT_E);
            cpt_vfor++;
          }

          if ((Lsv.mvt[ix] & (MVT_EH | MVT_OH)) == MVT_OH){
            UnsetMask(Lsv.mvt[ix], MVT_OH);
            SetMask(Lsv.mvt[ix], MVT_EH);
            cpt_vfor++;
          }

          if ((Lsv.mvt[ix] & (MVT_EB | MVT_OB)) == MVT_OB){
            UnsetMask(Lsv.mvt[ix], MVT_OB);
            SetMask(Lsv.mvt[ix], MVT_EB);
            cpt_vfor++;
          }
        }
        //if (!mvt_solid(ix)) Lsv.mvt[ix] = MVT_E;
        //if (!mvt_solid(ix+1)) Lsv.mvt[ix+1] = MVT_E;
        //if (!mvt_solid(ix+2)) Lsv.mvt[ix+2] = MVT_E;
      }
    }
  }
#endif //MVT_VFOR

#ifdef MVT_GFOR
  // forcage global du flux
  int cpt_for = 0;
  int nb_mvt_for = Lsv.nfc * intensite_gfor;

  //LogPrintf ("forcage sur %d cellules MVT\n", nb_mvt_for);

  while (nb_mvt_for > 0){
    //ix = drand48()*Lsv.sz;
    ix = Lsv.pmin*Lsv.lxy + drand48()*((Lsv.pmax-Lsv.pmin+1)*Lsv.lxy);
    if (Lsv.mvt[ix] & MVT_SOLID) continue;
    if ((Lsv.mvt[ix] & (MVT_E | MVT_O)) == MVT_O){
      UnsetMask(Lsv.mvt[ix], MVT_O);
      SetMask(Lsv.mvt[ix], MVT_E);
      cpt_for++;
    }

    if ((Lsv.mvt[ix] & (MVT_EH | MVT_OH)) == MVT_OH){
      UnsetMask(Lsv.mvt[ix], MVT_OH);
      SetMask(Lsv.mvt[ix], MVT_EH);
      cpt_for++;
    }

    if ((Lsv.mvt[ix] & (MVT_EB | MVT_OB)) == MVT_OB){
      UnsetMask(Lsv.mvt[ix], MVT_OB);
      SetMask(Lsv.mvt[ix], MVT_EB);
      cpt_for++;
    }
    nb_mvt_for--;
  }
  //LogPrintf ("forcage de %d particules MVT\n", cpt_for);
#endif
}

//#ifdef BROUILLAGE_MVT
void do_mixing()
{
  int i, y, z, y1, y2;
  int cix, cix1, cix2, mvt;

  for(z=Lsv.pmin; z<=Lsv.pmax; z++){
    for (i=0; i<(Lsv.ly/2); i++){

      //brouillage directionnel (bord gauche)
      y = 5+(Lsv.ly-10)*drand48();
      cix = z*Lsv.lxy+y*Lsv.lx+NB_MVT_EO;
      mvt = Lsv.mvt[cix] & (MVT_EB | MVT_EH);
      if (mvt == MVT_EB){
        UnsetMask(Lsv.mvt[cix], MVT_EB);
        SetMask(Lsv.mvt[cix], MVT_EH);
      }
      else if (mvt == MVT_EH){
        UnsetMask(Lsv.mvt[cix], MVT_EH);
        SetMask(Lsv.mvt[cix], MVT_EB);
      }

      //brouillage positionnel (bord gauche)
      y1 = 5+(Lsv.ly-10)*drand48();
      y2 = 5+(Lsv.ly-10)*drand48();
      cix1 = z*Lsv.lxy+y1*Lsv.lx+NB_MVT_EO;
      cix2 = z*Lsv.lxy+y2*Lsv.lx+NB_MVT_EO;
      if (!mvt_solid(cix1) && !mvt_solid(cix2)){
        //permutation
        mvt = Lsv.mvt[cix1];
        Lsv.mvt[cix1] = Lsv.mvt[cix2];
        Lsv.mvt[cix2] = mvt;
      }
    }
  }
}
//#endif


void compute_vel(char flag_interp)
{
  static int cpt=0;
  //static int cpt_dump=0;
  int i, j, k, ix, cpt1;
  int *pvx, *pvy;
  float fluid_ratio;

  /// averaging over neighboring space by using fixed windows
  //LogPrintf("compute_vel: moyennage dans l'espace\n");
  int iv, jv, imax, jmax;
  int nb_mvt_fluide;

#ifdef OPENMP
#pragma omp parallel for private(i,j,k,iv,jv,ix,pvx,pvy,imax,jmax,nb_mvt_fluide,fluid_ratio)
#endif
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ //profondeur
    pvx = Vsv.velx_time + Vsv.sz*cpt + Vsv.lxy*k;
    pvy = Vsv.vely_time + Vsv.sz*cpt + Vsv.lxy*k;
    for(jv=0; jv<Vsv.ly; jv++){ //hauteur
      for(iv=0; iv<Vsv.lx; iv++, pvx++, pvy++){ //largeur
        // moyenne des vitesses dans un rectangle de taille VSTEP_Y x VSTEP_X
        ix = k*Lsv.lxy + jv*Lsv.lx*VSTEP_Y + iv*VSTEP_X;
        if (jv < Vsv.ly-1) jmax = VSTEP_Y; else jmax = Lsv.ly-jv*VSTEP_Y;
        if (iv < Vsv.lx-1) imax = VSTEP_X; else imax = Lsv.lx-iv*VSTEP_X;
        *pvx = *pvy = 0;
        nb_mvt_fluide = 0;
        for(j=0; j<jmax; j++, ix+=Lsv.lx-imax){ //hauteur
          for(i=0; i<imax; i++, ix++){ //largeur
            *pvx += CelVelx[Lsv.mvt[ix]];
            *pvy += CelVely[Lsv.mvt[ix]];

            if (!(Lsv.mvt[ix] & MVT_SOLID)){
              nb_mvt_fluide++;
            }
          }
        }
        //moyennage sur les cellules fluides uniquement
        fluid_ratio = (float)nb_mvt_fluide/(imax*jmax);
        if (fluid_ratio < 0.2) fluid_ratio = 0.2;
        *pvx = (*pvx)/fluid_ratio;
        *pvy = (*pvy)/fluid_ratio;
      }
    }
  }

  /// averaging over time by using a circular buffer
  //LogPrintf("compute_vel: moyennage dans le temps\n");
  cpt1 = cpt+1;
  if (cpt1==VSTEP_TIME) cpt1 = 0;
  if (VSTEP_TIME > 1){
    for(i=0; i<Vsv.sz; i++){
      Vsv.velx[i] += Vsv.velx_time[cpt*Vsv.sz+i] - Vsv.velx_time[cpt1*Vsv.sz+i];
      Vsv.vely[i] += Vsv.vely_time[cpt*Vsv.sz+i] - Vsv.vely_time[cpt1*Vsv.sz+i];
    }
    //if (Vsv.velx[0]) {LogPrintf("Vsv.velx[0]=%d\tcpt=%d\n", Vsv.velx[0], cpt); }
  }

  /// computing mean and max velocity values over all space
  /*if (lgca_ready)*/{
    float vel_max = 0;
    int vel_size_2d = Vsv.sz/Lsv.npmax;
    int vx, vy;
    float vel, vel_sum;
    vel_sum = 0;
    Vsv.maxvel = 0;
#ifdef OPENMP
#pragma omp parallel for private(i,k,vx,vy,vel,vel_max) reduction(+:vel_sum)
#endif
    for(k=Lsv.pmin; k<=Lsv.pmax; k++){
      vel_max=0;
      for(i=0; i<vel_size_2d; i++){
        vx = Vsv.velx[i+k*vel_size_2d];
        vy = Vsv.vely[i+k*vel_size_2d];
        vel = sqrt(vx*vx + vy*vy);
        vel_sum += vel;
        if (vel_max<vel) vel_max = vel;
      }
#pragma omp critical
      {
        if (Vsv.maxvel<vel_max) Vsv.maxvel = vel_max;
      }
    }
    Vsv.meanvel = vel_sum/(vel_size_2d*(Lsv.pmax-Lsv.pmin+1)); //Vsv.sz;
  }

  if (flag_interp) compute_vel_interp();

  cpt = cpt1;
}

/// interpolation of the velocity
void compute_vel_interp()
{
  int i, j, k;
  int ix, ix0, ix1;

  if (!Vsv.velx_interp){
    AllocMemoryPrint("Vsv.velx_interp", Vsv.velx_interp, float, HLD);
    ResetMemory(Vsv.velx_interp, float, HLD);
  }
  if (!Vsv.vely_interp){
    AllocMemoryPrint("Vsv.vely_interp", Vsv.vely_interp, float, HLD);
    ResetMemory(Vsv.vely_interp, float, HLD);
    PrintTotalMemory();
  }

  int iv, jv;
  int *pvx, *pvy;
  int Velx00, Velx01, Velx10, Velx11;
  int Vely00, Vely01, Vely10, Vely11;
  int a00, a01, a10, a11;
  int imax = VSTEP/NB_MVT_EO;

  /// interpolation in each vertical plane
  //LogPrintf("interpolation dans chaque plan de collisions\n");
#ifdef OPENMP
#pragma omp parallel for private(i,j,k,iv,jv,ix,pvx,pvy,Velx00,Velx01,Velx10,Velx11,Vely00,Vely01,Vely10,Vely11,a00,a01,a10,a11)
#endif
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ //profondeur
    for(jv=0; jv<Vsv.ly-1; jv++){ //hauteur
      pvx = Vsv.velx+k*Vsv.lxy+jv*Vsv.lx;
      pvy = Vsv.vely+k*Vsv.lxy+jv*Vsv.lx;
      Velx00 = Velx01 = Velx10 = Velx11 = 0;
      Vely00 = Vely01 = Vely10 = Vely11 = 0;
      for(iv=0; iv<Vsv.lx-2; iv++, pvx++, pvy++){ //largeur
        // interpolation des vitesses dans un rectangle de taille (imax)xVSTEP
        ix = (LN+k*DIST_MVT_NS)*HL + jv*L*VSTEP + iv*imax;// + (VSTEP>>1)*L + (VSTEP>>1);
        Velx00 = *pvx;
        Velx10 = *(pvx+Vsv.lx);
        Vely00 = *pvy;
        Vely10 = *(pvy+Vsv.lx);
        if (imax > 1){
          Velx01 = *(pvx+1);
          Velx11 = *(pvx+Vsv.lx+1);
          Vely01 = *(pvy+1);
          Vely11 = *(pvy+Vsv.lx+1);
        }
        for(j=0; j<VSTEP; j++, ix+=L-imax){ //hauteur
          if (imax > 1){ //interpolation entre les 4 sommets du rectangle englobant
            for(i=0; i<imax; i++, ix++){ //largeur
              a00 = (VSTEP-j)*(imax-i);
              a01 = (VSTEP-j)*i;
              a10 = j*(imax-i);
              a11 = j*i;
              Vsv.velx_interp[ix] = (float)(Velx00*a00 + Velx01*a01 + Velx10*a10 + Velx11*a11)/(VSTEP*imax);
              Vsv.vely_interp[ix] = (float)(Vely00*a00 + Vely01*a01 + Vely10*a10 + Vely11*a11)/(VSTEP*imax);
            }
          }
          else{ //interpolation verticale entre 2 sommets
            Vsv.velx_interp[ix] = (float)(Velx00*(VSTEP-j) + Velx10*j)/VSTEP;
            Vsv.vely_interp[ix] = (float)(Vely00*(VSTEP-j) + Vely10*j)/VSTEP;
            ix+=imax;
          }
        }
      }

      // interpolation a cote du bord droit
      // on repete la derniere valeur interpolee
      // c'est plus facile et on reduit les effets de bord ...
      ix = (LN+k*DIST_MVT_NS)*HL + jv*L*VSTEP + iv*imax;// + (VSTEP>>1)*L + (VSTEP>>1);
      for(j=0; j<VSTEP; j++, ix+=L){ //hauteur
        for(i=0; i<L-iv*imax; i++){ //largeur
          Vsv.velx_interp[ix+i] = Vsv.velx_interp[ix-1];
          Vsv.vely_interp[ix+i] = Vsv.vely_interp[ix-1];
        }
      }
    }
  }


#if (DIST_MVT_NS > 1)
  float fz0, fz1;
  int kv, kvmax;

  /// north-south interpolation between all vertical planes
  //LogPrintf("interpolation nord-sud dans les plans intermediaires\n");
#ifdef OPENMP
#pragma omp parallel for private(i,j,k,kv,kvmax,ix,ix0,ix1,fz0,fz1)
#endif
  for(k=Lsv.pmin; k<=Lsv.pmax; k++){ /// vertical planes (lattice gas)
    kvmax = (k<Lsv.pmax) ? DIST_MVT_NS : LNS-Lsv.pmax*DIST_MVT_NS;
    for(kv=1; kv<kvmax; kv++){ /// intermediate planes (interpolation)
      ix = (LN+k*DIST_MVT_NS+kv)*HL;
      ix0 = (LN+k*DIST_MVT_NS)*HL;
      //ix1 = (LN+(k+1)*DIST_MVT_NS)*HL;
      ix1 = (k<Lsv.pmax) ? (LN+(k+1)*DIST_MVT_NS)*HL : pbc_mode ? LN*HL : ix0;
      fz1 = (float)kv/kvmax;
      fz0 = 1.0-fz1;
      for(j=0; j<H; j++){
        for(i=0; i<L; i++, ix++, ix0++, ix1++){
          if (rot_map && OutOfSpace(i,LN+k*DIST_MVT_NS+kv)) continue; /// optimization: no interpolation outside the rotating space
          Vsv.velx_interp[ix] = Vsv.velx_interp[ix0]*fz0 + Vsv.velx_interp[ix1]*fz1;
          Vsv.vely_interp[ix] = Vsv.vely_interp[ix0]*fz0 + Vsv.vely_interp[ix1]*fz1;
        }
      }
    }
  }
#endif
}

void dump_densite()
{
  static int cpt = 0;
  FILE *fp;
  int ix, n;
  //int nb_cel_fluide_dump = 0;

  //recalcul nb_mvt, nb_mvt_sol et Lsv.nfc
  int nb_mvt = 0; //nombre de particules mobiles
  int nb_mvt_sol = 0; //nombre de particules mobiles localisees dans une cellule solide
  //for (ix=0; ix<Lsv.sz; ix++){
  for (ix=Lsv.lxy*Lsv.pmin; ix<Lsv.lxy*(Lsv.pmax+1); ix++){
    n = NbMvtCell[Lsv.mvt[ix]];
    nb_mvt += n;
    if ((Lsv.mvt[ix] & MVT_SOLID)) nb_mvt_sol += n;
    //if (!(Lsv.mvt[ix] & MVT_SOLID)) nb_cel_fluide_dump++;
  }
  //Lsv.density = (float)nb_mvt / Lsv.nfc;
  Lsv.density = (float)(nb_mvt - nb_mvt_sol) / Lsv.nfc;

  //fichier densite
  fp = fopen("DENSI.log","a");
  if (!cpt) fprintf(fp,"\tmvt cells \tfluid nodes \tdensity \ttrapped cells\n");
  fprintf(fp,"%04d: \t%09ld \t%09ld \t%f \t%09ld\n", cpt++, (long)nb_mvt, Lsv.nfc, Lsv.density, (long)nb_mvt_sol);
  //fprintf(fp,"%03d : \t%09ld \t%09ld (%09ld) \t%f\n", cpt++, (long)nb_mvt, Lsv.nfc, nb_cel_fluide_dump, Lsv.density);
  fclose(fp);
}

void dump_vel()
{
  static int cpt = 0;
  static char start = 1;
  FILE *fp;

  if ((!Vsv.velx_interp) || (!Vsv.vely_interp)) {cpt++; return;}

#ifdef STABILITY_ANALYSIS
  //recalcul Vsv.maxvel et Vsv.meanvel  (calcul plus precis mais plus long, car on elimine les cellules solides lors du moyennage)
  int ix, ncel;
  float vx, vy, vv;
  Vsv.maxvel = 0.0;
  Vsv.meanvel = 0.0;
  //float Vsv.meanvel2 = 0.0;
  ncel = 0;
  for (ix=0; ix<HLD; ix++){
    vx = Vsv.velx_interp[ix];
    vy = Vsv.vely_interp[ix];
    vv = vx*vx+vy*vy;
    if (vv>Vsv.maxvel) Vsv.maxvel=vv;
    if (Phase[TE[ix].celltype] == FLUID) {Vsv.meanvel += vv; /*Vsv.meanvel2 += sqrt(vv);*/ ncel++;}
  }
  Vsv.maxvel=sqrt(Vsv.maxvel);
  Vsv.meanvel=sqrt(Vsv.meanvel/ncel); //approximation (en toute rigueur, il faut calculer la moyenne des racines carrees => plus de temps de calcul)
  //Vsv.meanvel2/=ncel; //valeur exacte
#endif

  //fichier VEL.log
  fp = fopen("VEL.log","a");
  if (start){
    start = 0;
    fprintf(fp,"      \t Vsv.maxvel \t Vsv.meanvel\n");
  }
  fprintf(fp,"%04d: \t  %04d  \t  %04d\n", cpt++, (int)Vsv.maxvel, (int)Vsv.meanvel);
  //fprintf(fp,"%03d : \t  %04d  \t  %04d  \t  %04d\n", cpt++, (int)Vsv.maxvel, (int)Vsv.meanvel, (int)Vsv.meanvel2);
  //fprintf(fp,"nb_mvt = %lu\n", nb_mvt);
  fclose(fp);
}

void dump_mvt(int cpt, int unit)
{
  static char filename[100];
  FILE *fp;

#ifdef LGCA_SAVE_BINARY
  // binary  file
  //sprintf(filename,"HPP_%d_%d_%d.bin",Lsv.np,Lsv.ly,Lsv.lx);
  if (unit == UNIT_COMP)
    sprintf(filename,"HPP%04d.bin", cpt);
  else
    sprintf(filename,"HPP%05d_t0.bin", cpt);

  fp = fopen(filename,"w");
  if ( ! fp ){
    ErrPrintf("ERROR: cannot open binary file %s\n", filename);
    exit(-4);
  }


  //while (CelMvt_lock) {LogPrintf("lock3\n"); };

  //CelMvt_lock = 1;

  fwrite(Lsv.mvt, sizeof(MvtField), Lsv.sz, fp);

  //CelMvt_lock = 0;

  fclose(fp);
#endif

  //free(buf);
  /*if (inter){
    char command[100];
    sprintf(command, "gzip %s", nom);
    system(command);
  }*/

  // fichier vitesses
//#if (VSTEP > 1)
#if 1
  if (Vsv.velx && Vsv.vely){
    if (unit == UNIT_COMP)
      sprintf(filename,"HPP_VEL%04d.bin", cpt, Lsv.np, Vsv.ly, Vsv.lx);
    else
      sprintf(filename,"HPP_VEL%05d.bin", cpt, Lsv.np, Vsv.ly, Vsv.lx);

    fp = fopen(filename,"w");
    if ( ! fp ){
      ErrPrintf("ERROR: cannot open VEL file %s\n", filename);
      exit(-4);
    }

    // header
    int step=5;
    fwrite(&Lsv.np, sizeof(int), 1, fp);
    fwrite(&Vsv.lx, sizeof(int), 1, fp);
    fwrite(&Vsv.ly, sizeof(int), 1, fp);
    fwrite(&step, sizeof(int), 1, fp);

    //compute_vel();

    // velocity data
    fwrite(Vsv.velx + Lsv.pmin*Vsv.lxy, sizeof(int), Lsv.np*Vsv.lxy, fp);
    fwrite(Vsv.vely + Lsv.pmin*Vsv.lxy, sizeof(int), Lsv.np*Vsv.lxy, fp);

    fclose(fp);
  }
#endif
}

void dump_collisions()
{
  FILE *fp;
  int mvt_in, mvt_out;

  /// save collision rules between moving particles

  fp = fopen("LGCA.log","w");

  fprintf(fp, "          IN            |          OUT\n");
  fprintf(fp, "E  W  B  T EB ET WB WT  |  E  W  B  T EB ET WB WT\n");
  fprintf(fp, "-------------------------------------------------\n");
  for(mvt_in=0; mvt_in<SIZE_MVT_FIELD; mvt_in+=4){
    mvt_out = Collisions[mvt_in];
    fprintf(fp, "%d  %d  %d  %d  %d  %d  %d  %d  |  %d  %d  %d  %d  %d  %d  %d  %d\n",
      (mvt_in & MVT_E)? 1 : 0, (mvt_in & MVT_O)? 1 : 0, (mvt_in & MVT_B)? 1 : 0, (mvt_in & MVT_H)? 1 : 0,
      (mvt_in & MVT_EB)? 1 : 0, (mvt_in & MVT_EH)? 1 : 0, (mvt_in & MVT_OB)? 1 : 0, (mvt_in & MVT_OH)? 1 : 0,
      (mvt_out & MVT_E)? 1 : 0, (mvt_out & MVT_O)? 1 : 0, (mvt_out & MVT_B)? 1 : 0, (mvt_out & MVT_H)? 1 : 0,
      (mvt_out & MVT_EB)? 1 : 0, (mvt_out & MVT_EH)? 1 : 0, (mvt_out & MVT_OB)? 1 : 0, (mvt_out & MVT_OH)? 1 : 0);
  }

#ifdef SOLID_SLIP
  //dump des regles de collision solide-fluide pour les particules lentes
  //selon la configuration du voisinage
  int slow[4] = {MVT_E, MVT_O, MVT_B, MVT_H};
  int mvt_sol_E, mvt_sol_O, mvt_sol_B, mvt_sol_H, mvt_sol, i;
  fprintf(fp, "\n   SOLID     |      IN      |      OUT\n");
  fprintf(fp, "E  O  B  H   | E  O  B  H   | E  O  B  H\n");
  for (mvt_sol=0; mvt_sol<16; mvt_sol++){
    mvt_sol_E = (mvt_sol & (MVT_E >> 2))? 1 : 0;
    mvt_sol_O = (mvt_sol & (MVT_O >> 2))? 1 : 0;
    mvt_sol_B = (mvt_sol & (MVT_B >> 2))? 1 : 0;
    mvt_sol_H = (mvt_sol & (MVT_H >> 2))? 1 : 0;
    for (i=0; i<4; i++){
      mvt_out = Collisions_SolSlow[mvt_sol][slow[i]];
      //mvt_in = MVT_SOLID | (1 << (2+i));
      //mvt_out = Collisions[mvt_in];
      fprintf(fp, "% d  %d  %d  %d  |  %d  %d  %d  %d  |  %d  %d  %d  %d\n",
              mvt_sol_E, mvt_sol_O, mvt_sol_B, mvt_sol_H,
              (i==0)? 1 : 0, (i==1)? 1 : 0, (i==2)? 1 : 0, (i==3)? 1 : 0,
              (mvt_out & MVT_E)? 1 : 0, (mvt_out & MVT_O)? 1 : 0, (mvt_out & MVT_B)? 1 : 0, (mvt_out & MVT_H)? 1 : 0);
    }
  }

  //dump des regles de collision solide-fluide pour les particules rapides (en diagonale)
  //selon la configuration du voisinage
  int fast[4] = {MVT_EB, MVT_EH, MVT_OB, MVT_OH};
  int mvt_sol_EB, mvt_sol_OB, mvt_sol_EH, mvt_sol_OH;//, mvt_sol, i;
  fprintf(fp, "\n   SOLID     |      IN      |      OUT\n");
  fprintf(fp, "EB EH OB OH  | EB EH OB OH  | EB EH OB OH\n");
  for (mvt_sol=0; mvt_sol<16; mvt_sol++){
    mvt_sol_EB = (mvt_sol & (MVT_EB >> 6))? 1 : 0;
    mvt_sol_EH = (mvt_sol & (MVT_EH >> 6))? 1 : 0;
    mvt_sol_OB = (mvt_sol & (MVT_OB >> 6))? 1 : 0;
    mvt_sol_OH = (mvt_sol & (MVT_OH >> 6))? 1 : 0;
    for (i=0; i<4; i++){
      mvt_out = Collisions_SolFast[mvt_sol][fast[i]];
      //mvt_in = MVT_SOLID | (1 << (6+i));
      //mvt_out = Collisions[mvt_in];
      fprintf(fp, "% d  %d  %d  %d  |  %d  %d  %d  %d  |  %d  %d  %d  %d\n",
              mvt_sol_EB, mvt_sol_EH, mvt_sol_OB, mvt_sol_OH,
              (i==0)? 1 : 0, (i==1)? 1 : 0, (i==2)? 1 : 0, (i==3)? 1 : 0,
              (mvt_out & MVT_EB)? 1 : 0, (mvt_out & MVT_EH)? 1 : 0, (mvt_out & MVT_OB)? 1 : 0, (mvt_out & MVT_OH)? 1 : 0);
    }
  }
#endif
  fclose(fp);
}



#ifdef DUMP_SIGNATURE
void dump_signature_mvt()
{
  FILE *fp;
  int i;
  unsigned int sig, *aux;

  //calcul de la signature
  sig=0;
  aux = (unsigned int*)Lsv.mvt;
  for(i=0; i<Lsv.sz*sizeof(MvtField)/sizeof(unsigned int); i++, aux++) sig += (*aux); //sig = sig ^ (*aux);

  //dump
  fp = fopen("SIGN_HPP.log","a");
  if ( ! fp ){
    ErrPrintf("erreur ouverture fichier dump signature HPP\n");
    exit(-4);
  }
  fprintf(fp, "%08x : %08x\n", lgca_iter, sig);
  fclose(fp);

}
#endif

#endif //LGCA
